"""hosts_resolver module"""

import argparse
import contextlib
import ctypes
import ipaddress
import json
import logging
import os
import socket
import sys
import typing
from pathlib import Path


class _Hostent(ctypes.Structure):  # pylint: disable=too-few-public-methods
    """hostent netdb.h structure ctypes mapping"""

    _fields_ = [
        ("h_name", ctypes.c_char_p),
        ("h_aliases", ctypes.POINTER(ctypes.c_char_p)),
        ("h_addrtype", ctypes.c_int),
        ("h_length", ctypes.c_int),
        ("h_addr_list", ctypes.POINTER(ctypes.POINTER(ctypes.c_ubyte))),
    ]


_IPAddress = typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

_libc = ctypes.CDLL("libc.so.6")

# From <netdb.h> :
## struct hostent *gethostent(void);
_gethostent = _libc.gethostent
_gethostent.restype = ctypes.POINTER(_Hostent)
_gethostent.argtypes = ()

# void endhostent(void);
_endhostent = _libc.endhostent
_endhostent.restype = None
_endhostent.argtypes = ()

_logger = logging.getLogger(Path(__file__).stem)


class _HostsResolver:
    """
    _HostsResolver context manager definition.

    Use this CM to resolve name using /etc/hosts entries without any DNS request or subprocess call.

    Known limitations:
      - `gethostent` is not reentrant (see GETHOSTBYNAME(3))
      - `gethostent` is not thread-safe (see GETHOSTBYNAME(3))
      - glibc implementations does not support IPv6 addresses (see GETHOSTBYNAME(3))
      - behavior is **undefined** when using multiple context managers simultaneously
    """

    def __init__(self) -> None:
        self._cached_aliases: typing.Dict[str, _IPAddress] = {}

    def __enter__(self) -> "_HostsResolver":
        return self

    def __exit__(self, _, __, ___):
        _endhostent()

    def resolve(self, name: typing.Union[str, _IPAddress]) -> _IPAddress:
        """Try to lazily resolve `name` using /etc/hosts entries. Results are cached"""
        # don't try to resolve IP addresses for idempotency !
        with contextlib.suppress(ValueError):
            return ipaddress.ip_address(name)

        while True:
            # fetch result from cache (if present)
            with contextlib.suppress(KeyError):
                return self._cached_aliases[name]  # type: ignore[index]

            # fetch "next" hosts entry
            entry_ptr = ctypes.cast(_gethostent(), ctypes.POINTER(_Hostent))
            if not entry_ptr:
                _logger.debug("no more hosts entry")
                break

            # dereference structure pointer
            entry = entry_ptr.contents

            # decode entry "name"
            aliases = [entry.h_name.decode("ascii")]

            _logger.debug("got new hosts entry: %s", aliases[0])

            # decode entry "aliases"
            i = 0
            h_aliases = ctypes.cast(entry.h_aliases, ctypes.POINTER(ctypes.c_char_p))
            while h_aliases[i]:
                aliases.append(h_aliases[i].decode("ascii"))
                _logger.debug("found new alias for %s: %s", aliases[0], aliases[-1])
                i += 1

            # decode "network addresses" from raw bytes array
            addresses = ctypes.cast(
                entry.h_addr_list,
                ctypes.POINTER(ctypes.POINTER(ctypes.c_ubyte * entry.h_length)),
            )
            address = ipaddress.ip_address(bytes(addresses[0].contents))

            # populate cache with all entry aliases
            for alias in aliases:
                _logger.debug("caching %s as %s...", address, alias)
                self._cached_aliases[alias] = address

        raise socket.gaierror(f"could not resolve: {name}")


@contextlib.contextmanager
def get_hosts_resolver() -> typing.Generator[_HostsResolver, None, None]:
    """Return an `_HostsResolver` instance to use as a context manager"""
    with _HostsResolver() as hosts_resolver:
        yield hosts_resolver


def main() -> typing.NoReturn:
    """Main entry point"""
    parser = argparse.ArgumentParser(
        description="Zero-dependency Python module to resolve names using /etc/hosts alone"
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s 1.0.0",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="enable debug mode",
    )
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="output results as JSON",
    )
    parser.add_argument(
        "names",
        nargs="+",
        help="names to resolve using /etc/hosts",
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    results: typing.Dict[str, typing.Optional[str]] = {}
    with get_hosts_resolver() as hosts_resolver:
        for name in args.names:
            try:
                results[name] = str(hosts_resolver.resolve(name))
            except socket.gaierror:
                # if JSON output is disabled, exit fast to prevent inconsistent output
                if not args.json:
                    raise

                _logger.exception("resolution failed for %s", name)
                results[name] = None

    if args.json:
        print(json.dumps(results, indent=4 if args.debug else None))
    else:
        print(os.linesep.join(results.values()))  # type: ignore[arg-type]

    # exit with error if (at least) a resolution failed
    sys.exit(not all(results.values()))


if __name__ == "__main__":
    main()
